import dataclasses
import re
from datetime import date, datetime, timedelta
from typing import Any, Dict, List, Union


CAMEL_TO_SNAKE = re.compile(
    r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z])|(?<=[a-zA-Z])[0-9])"
)


def camel_to_snake(camel_str: str) -> str:
    snake_str = CAMEL_TO_SNAKE.sub(r"_\1", camel_str)
    return snake_str.lower()


def camel_to_snake_dict(camel_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Converts a dictionary's keys from camel case to snake case. This version
    handles nested dictionaries and lists.
    """
    snake_dict: Dict[str, Any] = {}
    for k, v in camel_dict.items():
        new_key = camel_to_snake(k)
        if isinstance(v, dict):
            snake_dict[new_key] = camel_to_snake_dict(v)
        elif isinstance(v, list):
            snake_dict[new_key] = [
                camel_to_snake_dict(i) if isinstance(i, dict) else i for i in v
            ]
        else:
            snake_dict[new_key] = v
    return snake_dict


def format_end_date(end: Union[date, str, None]) -> date:
    if end is None:
        end = date.today()
    elif isinstance(end, str):
        end = date.fromisoformat(end)
    return end


def date_range(date_: Union[date, str], days: int):
    date_ = date_ if isinstance(date_, date) else date.fromisoformat(date_)
    for day in range(days):
        yield date_ - timedelta(days=day)


def asdict(obj):
    if dataclasses.is_dataclass(obj):
        result = {}
        for field in dataclasses.fields(obj):
            value = getattr(obj, field.name)
            result[field.name] = asdict(value)
        return result

    if isinstance(obj, List):
        return [asdict(v) for v in obj]

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    return obj



from typing import ClassVar

from pydantic.dataclasses import dataclass

from ._base import Stats


BASE_PATH = "/usersummary-service/stats/steps"


@dataclass(frozen=True)
class DailySteps(Stats):
    total_steps: int
    total_distance: int
    step_goal: int

    _path: ClassVar[str] = f"{BASE_PATH}/daily/{{start}}/{{end}}"
    _page_size: ClassVar[int] = 28


@dataclass(frozen=True)
class WeeklySteps(Stats):
    total_steps: int
    average_steps: float
    average_distance: float
    total_distance: float
    wellness_data_days_count: int

    _path: ClassVar[str] = f"{BASE_PATH}/weekly/{{end}}/{{period}}"
    _page_size: ClassVar[int] = 52



from datetime import date
from typing import Dict, List

from pydantic.dataclasses import dataclass

from .. import http
from ..utils import camel_to_snake_dict


@dataclass
class PowerFormat:
    format_id: int
    format_key: str
    min_fraction: int
    max_fraction: int
    grouping_used: bool
    display_format: str | None


@dataclass
class FirstDayOfWeek:
    day_id: int
    day_name: str
    sort_order: int
    is_possible_first_day: bool


@dataclass
class WeatherLocation:
    use_fixed_location: bool
    latitude: float
    longitude: float
    location_name: str
    iso_country_code: str
    postal_code: str


@dataclass
class UserData:
    gender: str
    weight: float
    height: float
    time_format: str
    birth_date: date
    measurement_system: str
    activity_level: str | None
    handedness: str
    power_format: PowerFormat
    heart_rate_format: PowerFormat
    first_day_of_week: FirstDayOfWeek
    vo_2_max_running: float | None
    vo_2_max_cycling: float | None
    lactate_threshold_speed: float | None
    lactate_threshold_heart_rate: float | None
    dive_number: int | None
    intensity_minutes_calc_method: str
    moderate_intensity_minutes_hr_zone: int
    vigorous_intensity_minutes_hr_zone: int
    hydration_measurement_unit: str
    hydration_containers: List[Dict[str, float | None]]
    hydration_auto_goal_enabled: bool
    firstbeat_max_stress_score: float | None
    firstbeat_cycling_lt_timestamp: int | None
    firstbeat_running_lt_timestamp: int | None
    threshold_heart_rate_auto_detected: bool
    ftp_auto_detected: bool | None
    training_status_paused_date: str | None
    weather_location: WeatherLocation | None
    golf_distance_unit: str
    golf_elevation_unit: str | None
    golf_speed_unit: str | None
    external_bottom_time: float | None


@dataclass
class UserSleep:
    sleep_time: int
    default_sleep_time: bool
    wake_time: int
    default_wake_time: bool


@dataclass
class UserSettings:
    id: int
    user_data: UserData
    user_sleep: UserSleep
    connect_date: str | None
    source_type: str | None

    @classmethod
    def get(cls, /, client: http.Client | None = None):
        client = client or http.client
        settings = client.connectapi(
            "/userprofile-service/userprofile/user-settings"
        )
        assert isinstance(settings, dict)
        return cls(**camel_to_snake_dict(settings))



import time

from garth.auth_tokens import OAuth2Token


def test_is_expired(oauth2_token: OAuth2Token):
    oauth2_token.expires_at = int(time.time() - 1)
    assert oauth2_token.expired is True


def test_refresh_is_expired(oauth2_token: OAuth2Token):
    oauth2_token.refresh_token_expires_at = int(time.time() - 1)
    assert oauth2_token.refresh_expired is True


def test_str(oauth2_token: OAuth2Token):
    assert str(oauth2_token) == "Bearer bar"



from typing import ClassVar

from pydantic.dataclasses import dataclass

from ._base import Stats


BASE_PATH = "/usersummary-service/stats/im"


@dataclass(frozen=True)
class DailyIntensityMinutes(Stats):
    weekly_goal: int
    moderate_value: int | None = None
    vigorous_value: int | None = None

    _path: ClassVar[str] = f"{BASE_PATH}/daily/{{start}}/{{end}}"
    _page_size: ClassVar[int] = 28


@dataclass(frozen=True)
class WeeklyIntensityMinutes(Stats):
    weekly_goal: int
    moderate_value: int | None = None
    vigorous_value: int | None = None

    _path: ClassVar[str] = f"{BASE_PATH}/weekly/{{start}}/{{end}}"
    _page_size: ClassVar[int] = 52



from typing import ClassVar

from pydantic.dataclasses import dataclass

from ._base import Stats


@dataclass(frozen=True)
class DailySleep(Stats):
    value: int | None

    _path: ClassVar[str] = (
        "/wellness-service/stats/daily/sleep/score/{start}/{end}"
    )
    _page_size: ClassVar[int] = 28



from .data import HRVData, SleepData
from .http import Client, client
from .stats import (
    DailyHRV,
    DailyIntensityMinutes,
    DailySleep,
    DailySteps,
    DailyStress,
    WeeklyIntensityMinutes,
    WeeklySteps,
    WeeklyStress,
)
from .users import UserProfile, UserSettings
from .version import __version__


__all__ = [
    "Client",
    "DailyHRV",
    "DailyIntensityMinutes",
    "DailySleep",
    "DailySteps",
    "DailyStress",
    "HRVData",
    "SleepData",
    "UserProfile",
    "UserSettings",
    "WeeklyIntensityMinutes",
    "WeeklySteps",
    "WeeklyStress",
    "__version__",
    "client",
    "configure",
    "connectapi",
    "download",
    "login",
    "resume",
    "save",
    "upload",
]

configure = client.configure
connectapi = client.connectapi
download = client.download
login = client.login
resume = client.load
save = client.dump
upload = client.upload



from datetime import date, datetime, timedelta
from typing import Any, ClassVar, cast

from pydantic.dataclasses import dataclass

from .. import http
from ..utils import camel_to_snake_dict, format_end_date


@dataclass(frozen=True)
class HRVBaseline:
    low_upper: int
    balanced_low: int
    balanced_upper: int
    marker_value: float | None


@dataclass(frozen=True)
class DailyHRV:
    calendar_date: date
    weekly_avg: int | None
    last_night_avg: int | None
    last_night_5_min_high: int | None
    baseline: HRVBaseline | None
    status: str
    feedback_phrase: str
    create_time_stamp: datetime

    _path: ClassVar[str] = "/hrv-service/hrv/daily/{start}/{end}"
    _page_size: ClassVar[int] = 28

    @classmethod
    def list(
        cls,
        end: date | str | None = None,
        period: int = 28,
        *,
        client: http.Client | None = None,
    ) -> list["DailyHRV"]:
        client = client or http.client
        end = format_end_date(end)

        # Paginate if period is greater than page size
        if period > cls._page_size:
            page = cls.list(end, cls._page_size, client=client)
            if not page:
                return []
            page = (
                cls.list(
                    end - timedelta(days=cls._page_size),
                    period - cls._page_size,
                    client=client,
                )
                + page
            )
            return page

        start = end - timedelta(days=period - 1)
        path = cls._path.format(start=start, end=end)
        response = client.connectapi(path)
        if response is None:
            return []
        daily_hrv = camel_to_snake_dict(response)["hrv_summaries"]
        daily_hrv = cast(list[dict[str, Any]], daily_hrv)
        return [cls(**hrv) for hrv in daily_hrv]



from dataclasses import dataclass

from requests import HTTPError


@dataclass(frozen=True)
class GarthException(Exception):
    """Base exception for all garth exceptions."""

    msg: str


@dataclass(frozen=True)
class GarthHTTPError(GarthException):
    error: HTTPError

    def __str__(self) -> str:
        return f"{self.msg}: {self.error}"



from typing import List

from pydantic.dataclasses import dataclass

from .. import http
from ..utils import camel_to_snake_dict


@dataclass(frozen=True)
class UserProfile:
    id: int
    profile_id: int
    garmin_guid: str
    display_name: str
    full_name: str
    user_name: str
    profile_image_uuid: str | None
    profile_image_url_large: str | None
    profile_image_url_medium: str | None
    profile_image_url_small: str | None
    location: str | None
    facebook_url: str | None
    twitter_url: str | None
    personal_website: str | None
    motivation: str | None
    bio: str | None
    primary_activity: str | None
    favorite_activity_types: List[str]
    running_training_speed: float
    cycling_training_speed: float
    favorite_cycling_activity_types: List[str]
    cycling_classification: str | None
    cycling_max_avg_power: float
    swimming_training_speed: float
    profile_visibility: str
    activity_start_visibility: str
    activity_map_visibility: str
    course_visibility: str
    activity_heart_rate_visibility: str
    activity_power_visibility: str
    badge_visibility: str
    show_age: bool
    show_weight: bool
    show_height: bool
    show_weight_class: bool
    show_age_range: bool
    show_gender: bool
    show_activity_class: bool
    show_vo_2_max: bool
    show_personal_records: bool
    show_last_12_months: bool
    show_lifetime_totals: bool
    show_upcoming_events: bool
    show_recent_favorites: bool
    show_recent_device: bool
    show_recent_gear: bool
    show_badges: bool
    other_activity: str | None
    other_primary_activity: str | None
    other_motivation: str | None
    user_roles: List[str]
    name_approved: bool
    user_profile_full_name: str
    make_golf_scorecards_private: bool
    allow_golf_live_scoring: bool
    allow_golf_scoring_by_connections: bool
    user_level: int
    user_point: int
    level_update_date: str
    level_is_viewed: bool
    level_point_threshold: int
    user_point_offset: int
    user_pro: bool

    @classmethod
    def get(cls, /, client: http.Client | None = None):
        client = client or http.client
        profile = client.connectapi("/userprofile-service/socialProfile")
        assert isinstance(profile, dict)
        return cls(**camel_to_snake_dict(profile))



import pytest

from garth import UserProfile, UserSettings
from garth.http import Client


@pytest.mark.vcr
def test_user_profile(authed_client: Client):
    profile = UserProfile.get(client=authed_client)
    assert profile.user_name


@pytest.mark.vcr
def test_user_setttings(authed_client: Client):
    settings = UserSettings.get(client=authed_client)
    assert settings.user_data



from datetime import date

import pytest

from garth import DailyIntensityMinutes, WeeklyIntensityMinutes
from garth.http import Client


@pytest.mark.vcr
def test_daily_intensity_minutes(authed_client: Client):
    end = date(2023, 7, 20)
    days = 20
    daily_im = DailyIntensityMinutes.list(end, days, client=authed_client)
    assert daily_im[-1].calendar_date == end
    assert len(daily_im) == days


@pytest.mark.vcr
def test_weekly_intensity_minutes(authed_client: Client):
    end = date(2023, 7, 20)
    weeks = 12
    weekly_im = WeeklyIntensityMinutes.list(end, weeks, client=authed_client)
    assert len(weekly_im) == weeks
    assert (
        weekly_im[-1].calendar_date.isocalendar()[
            1
        ]  # in python3.9+ [1] can be .week
        == end.isocalendar()[1]
    )



from datetime import date

import pytest

from garth import DailyHRV
from garth.http import Client


@pytest.mark.vcr
def test_daily_hrv(authed_client: Client):
    end = date(2023, 7, 20)
    days = 20
    daily_hrv = DailyHRV.list(end, days, client=authed_client)
    assert daily_hrv[-1].calendar_date == end
    assert len(daily_hrv) == days


@pytest.mark.vcr
def test_daily_hrv_paginate(authed_client: Client):
    end = date(2023, 7, 20)
    days = 40
    daily_hrv = DailyHRV.list(end, days, client=authed_client)
    assert daily_hrv[-1].calendar_date == end
    assert len(daily_hrv) == days


@pytest.mark.vcr
def test_daily_hrv_no_results(authed_client: Client):
    end = date(1990, 7, 20)
    daily_hrv = DailyHRV.list(end, client=authed_client)
    assert daily_hrv == []


@pytest.mark.vcr
def test_daily_hrv_paginate_no_results(authed_client: Client):
    end = date(1990, 7, 20)
    days = 40
    daily_hrv = DailyHRV.list(end, days, client=authed_client)
    assert daily_hrv == []



import tempfile
import time
from typing import Any, cast

import pytest
from requests.adapters import HTTPAdapter

from garth.auth_tokens import OAuth2Token
from garth.exc import GarthHTTPError
from garth.http import Client


def test_dump_and_load(authed_client: Client):
    with tempfile.TemporaryDirectory() as tempdir:
        authed_client.dump(tempdir)

        new_client = Client()
        new_client.load(tempdir)

        assert new_client.oauth1_token == authed_client.oauth1_token
        assert new_client.oauth2_token == authed_client.oauth2_token


def test_dumps_and_loads(authed_client: Client):
    s = authed_client.dumps()
    new_client = Client()
    new_client.loads(s)
    assert new_client.oauth1_token == authed_client.oauth1_token
    assert new_client.oauth2_token == authed_client.oauth2_token


def test_configure_oauth2_token(client: Client, oauth2_token: OAuth2Token):
    assert client.oauth2_token is None
    client.configure(oauth2_token=oauth2_token)
    assert client.oauth2_token == oauth2_token


def test_configure_domain(client: Client):
    assert client.domain == "garmin.com"
    client.configure(domain="garmin.cn")
    assert client.domain == "garmin.cn"


def test_configure_proxies(client: Client):
    assert client.sess.proxies == {}
    proxy = {"https": "http://localhost:8888"}
    client.configure(proxies=proxy)
    assert client.sess.proxies["https"] == proxy["https"]


def test_configure_ssl_verify(client: Client):
    assert client.sess.verify is True
    client.configure(ssl_verify=False)
    assert client.sess.verify is False


def test_configure_timeout(client: Client):
    assert client.timeout == 10
    client.configure(timeout=99)
    assert client.timeout == 99


def test_configure_retry(client: Client):
    assert client.retries == 3
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.total == client.retries

    client.configure(retries=99)
    assert client.retries == 99
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.total == 99


def test_configure_status_forcelist(client: Client):
    assert client.status_forcelist == (408, 429, 500, 502, 503, 504)
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.status_forcelist == client.status_forcelist

    client.configure(status_forcelist=(200, 201, 202))
    assert client.status_forcelist == (200, 201, 202)
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.status_forcelist == client.status_forcelist


def test_configure_backoff_factor(client: Client):
    assert client.backoff_factor == 0.5
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.backoff_factor == client.backoff_factor

    client.configure(backoff_factor=0.99)
    assert client.backoff_factor == 0.99
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.max_retries.backoff_factor == client.backoff_factor


def test_configure_pool_maxsize(client: Client):
    assert client.pool_maxsize == 10
    client.configure(pool_maxsize=99)
    assert client.pool_maxsize == 99
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert adapter.poolmanager.connection_pool_kw["maxsize"] == 99


def test_configure_pool_connections(client: Client):
    client.configure(pool_connections=99)
    assert client.pool_connections == 99
    adapter = client.sess.adapters["https://"]
    assert isinstance(adapter, HTTPAdapter)
    assert (
        getattr(adapter, "_pool_connections", None) == 99
    ), "Pool connections not properly configured"


@pytest.mark.vcr
def test_client_request(client: Client):
    resp = client.request("GET", "connect", "/")
    assert resp.ok

    with pytest.raises(GarthHTTPError) as e:
        client.request("GET", "connectapi", "/")
    assert "404" in str(e.value)


@pytest.mark.vcr
def test_login_success_mfa(monkeypatch, client: Client):
    def mock_input(_):
        return "327751"

    monkeypatch.setattr("builtins.input", mock_input)

    assert client.oauth1_token is None
    assert client.oauth2_token is None
    client.login("user@example.com", "correct_password")
    assert client.oauth1_token
    assert client.oauth2_token


@pytest.mark.vcr
def test_username(authed_client: Client):
    assert authed_client._social_profile is None
    assert authed_client.username
    assert authed_client._social_profile


@pytest.mark.vcr
def test_connectapi(authed_client: Client):
    stress = cast(
        list[dict[str, Any]],
        authed_client.connectapi(
            "/usersummary-service/stats/stress/daily/2023-07-21/2023-07-21"
        ),
    )
    assert stress
    assert isinstance(stress, list)
    assert len(stress) == 1
    assert stress[0]["calendarDate"] == "2023-07-21"
    assert list(stress[0]["values"].keys()) == [
        "highStressDuration",
        "lowStressDuration",
        "overallStressLevel",
        "restStressDuration",
        "mediumStressDuration",
    ]


@pytest.mark.vcr
def test_refresh_oauth2_token(authed_client: Client):
    assert authed_client.oauth2_token
    authed_client.oauth2_token.expires_at = int(time.time())
    assert authed_client.oauth2_token.expired
    profile = authed_client.connectapi("/userprofile-service/socialProfile")
    assert profile
    assert isinstance(profile, dict)
    assert profile["userName"]


@pytest.mark.vcr
def test_download(authed_client: Client):
    downloaded = authed_client.download(
        "/download-service/files/activity/11998957007"
    )
    assert downloaded
    zip_magic_number = b"\x50\x4b\x03\x04"
    assert downloaded[:4] == zip_magic_number


@pytest.mark.vcr
def test_upload(authed_client: Client):
    fpath = "tests/12129115726_ACTIVITY.fit"
    with open(fpath, "rb") as f:
        uploaded = authed_client.upload(f)
    assert uploaded


@pytest.mark.vcr
def test_delete(authed_client: Client):
    activity_id = "12135235656"
    path = f"/activity-service/activity/{activity_id}"
    assert authed_client.connectapi(path)
    authed_client.delete(
        "connectapi",
        path,
        api=True,
    )
    with pytest.raises(GarthHTTPError) as e:
        authed_client.connectapi(path)
    assert "404" in str(e.value)


@pytest.mark.vcr
def test_put(authed_client: Client):
    data = [
        {
            "changeState": "CHANGED",
            "trainingMethod": "HR_RESERVE",
            "lactateThresholdHeartRateUsed": 170,
            "maxHeartRateUsed": 185,
            "restingHrAutoUpdateUsed": False,
            "sport": "DEFAULT",
            "zone1Floor": 130,
            "zone2Floor": 140,
            "zone3Floor": 150,
            "zone4Floor": 160,
            "zone5Floor": 170,
        }
    ]
    path = "/biometric-service/heartRateZones"
    authed_client.put(
        "connectapi",
        path,
        api=True,
        json=data,
    )
    assert authed_client.connectapi(path)



import asyncio
import re
import time
from typing import Callable, Dict, Tuple
from urllib.parse import parse_qs

import requests
from requests import Session
from requests_oauthlib import OAuth1Session

from . import http
from .auth_tokens import OAuth1Token, OAuth2Token
from .exc import GarthException


CSRF_RE = re.compile(r'name="_csrf"\s+value="(.+?)"')
TITLE_RE = re.compile(r"<title>(.+?)</title>")
OAUTH_CONSUMER_URL = "https://thegarth.s3.amazonaws.com/oauth_consumer.json"
OAUTH_CONSUMER: Dict[str, str] = {}
USER_AGENT = {"User-Agent": "com.garmin.android.apps.connectmobile"}


class GarminOAuth1Session(OAuth1Session):
    def __init__(
        self,
        /,
        parent: Session | None = None,
        **kwargs,
    ):
        global OAUTH_CONSUMER
        if not OAUTH_CONSUMER:
            OAUTH_CONSUMER = requests.get(OAUTH_CONSUMER_URL).json()
        super().__init__(
            OAUTH_CONSUMER["consumer_key"],
            OAUTH_CONSUMER["consumer_secret"],
            **kwargs,
        )
        if parent is not None:
            self.mount("https://", parent.adapters["https://"])
            self.proxies = parent.proxies
            self.verify = parent.verify


def login(
    email: str,
    password: str,
    /,
    client: "http.Client | None" = None,
    prompt_mfa: Callable = lambda: input("MFA code: "),
) -> Tuple[OAuth1Token, OAuth2Token]:
    client = client or http.client

    # Define params based on domain
    SSO = f"https://sso.{client.domain}/sso"
    SSO_EMBED = f"{SSO}/embed"
    SSO_EMBED_PARAMS = dict(
        id="gauth-widget",
        embedWidget="true",
        gauthHost=SSO,
    )
    SIGNIN_PARAMS = {
        **SSO_EMBED_PARAMS,
        **dict(
            gauthHost=SSO_EMBED,
            service=SSO_EMBED,
            source=SSO_EMBED,
            redirectAfterAccountLoginUrl=SSO_EMBED,
            redirectAfterAccountCreationUrl=SSO_EMBED,
        ),
    }

    # Set cookies
    client.get("sso", "/sso/embed", params=SSO_EMBED_PARAMS)

    # Get CSRF token
    client.get(
        "sso",
        "/sso/signin",
        params=SIGNIN_PARAMS,
        referrer=True,
    )
    csrf_token = get_csrf_token(client.last_resp.text)

    # Submit login form with email and password
    client.post(
        "sso",
        "/sso/signin",
        params=SIGNIN_PARAMS,
        referrer=True,
        data=dict(
            username=email,
            password=password,
            embed="true",
            _csrf=csrf_token,
        ),
    )
    title = get_title(client.last_resp.text)

    # Handle MFA
    if "MFA" in title:
        handle_mfa(client, SIGNIN_PARAMS, prompt_mfa)
        title = get_title(client.last_resp.text)

    assert title == "Success", f"Unexpected title: {title}"

    # Parse ticket
    m = re.search(r'embed\?ticket=([^"]+)"', client.last_resp.text)
    assert m, "Couldn't find ticket in response"
    ticket = m.group(1)

    oauth1 = get_oauth1_token(ticket, client)
    oauth2 = exchange(oauth1, client)

    return oauth1, oauth2


def get_oauth1_token(ticket: str, client: "http.Client") -> OAuth1Token:
    sess = GarminOAuth1Session(parent=client.sess)
    resp = sess.get(
        f"https://connectapi.{client.domain}/oauth-service/oauth/"
        f"preauthorized?ticket={ticket}&login-url="
        f"https://sso.{client.domain}/sso/embed&accepts-mfa-tokens=true",
        headers=USER_AGENT,
        timeout=client.timeout,
    )
    resp.raise_for_status()
    parsed = parse_qs(resp.text)
    token = {k: v[0] for k, v in parsed.items()}
    return OAuth1Token(domain=client.domain, **token)  # type: ignore


def exchange(oauth1: OAuth1Token, client: "http.Client") -> OAuth2Token:
    sess = GarminOAuth1Session(
        resource_owner_key=oauth1.oauth_token,
        resource_owner_secret=oauth1.oauth_token_secret,
        parent=client.sess,
    )
    data = dict(mfa_token=oauth1.mfa_token) if oauth1.mfa_token else {}
    token = sess.post(
        f"https://connectapi.{client.domain}/oauth-service/oauth/exchange/user/2.0",
        headers={
            **USER_AGENT,
            **{"Content-Type": "application/x-www-form-urlencoded"},
        },
        data=data,
        timeout=client.timeout,
    ).json()
    return OAuth2Token(**set_expirations(token))


def handle_mfa(
    client: "http.Client", signin_params: dict, prompt_mfa: Callable
) -> None:
    csrf_token = get_csrf_token(client.last_resp.text)
    if asyncio.iscoroutinefunction(prompt_mfa):
        mfa_code = asyncio.run(prompt_mfa())
    else:
        mfa_code = prompt_mfa()
    client.post(
        "sso",
        "/sso/verifyMFA/loginEnterMfaCode",
        params=signin_params,
        referrer=True,
        data={
            "mfa-code": mfa_code,
            "embed": "true",
            "_csrf": csrf_token,
            "fromPage": "setupEnterMfaCode",
        },
    )


def set_expirations(token: dict) -> dict:
    token["expires_at"] = int(time.time() + token["expires_in"])
    token["refresh_token_expires_at"] = int(
        time.time() + token["refresh_token_expires_in"]
    )
    return token


def get_csrf_token(html: str) -> str:
    m = CSRF_RE.search(html)
    if not m:
        raise GarthException("Couldn't find CSRF token")
    return m.group(1)


def get_title(html: str) -> str:
    m = TITLE_RE.search(html)
    if not m:
        raise GarthException("Couldn't find title")
    return m.group(1)
